@@ -1,7 +1,7 @@
 # ABSTRACT: take structured address data and format it according to the various global/country rules
 
 package Geo::Address::Formatter;
-$Geo::Address::Formatter::VERSION = '1.51';
+$Geo::Address::Formatter::VERSION = '1.52';
 use strict;
 use warnings;
 
@@ -103,7 +103,6 @@ sub format_address {
     if ($cc){
 	$rh_components->{country_code} = $cc;
     } 
-
     #warn Dumper $rh_components;
 
     # set the aliases, unless this would overwrite something
@@ -116,7 +115,6 @@ sub format_address {
                 $rh_components->{$alias};
         }
     }
-
     $self->_sanity_cleaning($rh_components); 
 
     # determine the template
@@ -137,7 +135,6 @@ sub format_address {
         }
         # no fallback
     }
-    $template_text =~ s/\n/, /sg;
 
     # clean up the components
     $self->_fix_country($rh_components);
@@ -158,7 +155,8 @@ sub format_address {
     my $compiled_template = $THT_cache{$template_text};
 
     # render it
-    my $text = $self->_render_template($compiled_template, $rh_components);
+    my $text;
+    $text = $self->_render_template($compiled_template, $rh_components);
     $text = $self->_postformat($text,$rh_config->{postformat_replace});
     $text = $self->_clean($text);
     return $text;
@@ -171,7 +169,7 @@ sub _postformat {
     my $text_orig = $text; # keep a copy
 
     # remove duplicates
-    my @before_pieces = split(/,/, $text);
+    my @before_pieces = split(/, /, $text);
     my %seen;
     my @after_pieces;
     foreach my $piece (@before_pieces){
@@ -229,7 +227,6 @@ sub _sanity_cleaning {
     return;
 }
 
-
 sub _minimal_components {
     my $self = shift;
     my $rh_components = shift || return;
@@ -323,12 +320,11 @@ sub _determine_country_code {
     return;
 }
 
-# sets and returns a state code
+# hacks for bad country data
 sub _fix_country {
     my $self          = shift;
     my $rh_components = shift || return;
 
-    #warn Dumper $rh_components;
     # is the country a number?
     # if so, and there is a state, use state as country
     if (defined($rh_components->{country})){
@@ -339,7 +335,6 @@ sub _fix_country {
 	    }
         }
     }
-
     return;
 }
 
@@ -400,21 +395,55 @@ sub _apply_replacements {
 }
 
 # " abc,,def , ghi " => 'abc, def, ghi'
+
 sub _clean {
     my $self = shift;
-    my $out  = shift // '';
+    my $out  = shift // return undef;
+    #warn "entering _clean \n$out";
+
     $out =~ s/[\},\s]+$//;
     $out =~ s/^[,\s]+//;
 
     $out =~ s/,\s*,/, /g; # multiple commas to one   
-    $out =~ s/\s+,\s+/, /g; # one space behind comma
+    $out =~ s/\h+,\h+/, /g; # one horiz whitespace behind comma
+    $out =~ s/\h\h+/ /g;  # multiple horiz whitespace to one
+    $out =~ s/\h\n/\n/g;  # horiz whitespace, newline to newline
+    $out =~ s/\n,/\n/g;   # newline comma to just newline
+    $out =~ s/,,+/,/g;    # multiple commas to one
+    $out =~ s/,\n/\n/g;   # comma newline to just newline
+    $out =~ s/\n\h+/\n/g; # newline plus space to newline
+    $out =~ s/\n\n+/\n/g; # multiple newline to one
 
-    $out =~ s/\s\s+/ /g; # multiple whitespace to one
-    $out =~ s/,,+/,/g; # multiple commas to one
+    # final dedupe across and within lines
+    my @before_pieces = split(/\n/, $out);
+    my %seen_lines;
+    my @after_pieces;
+    foreach my $line (@before_pieces){
+        $line =~s/^\h+//g;
+        $line =~s/\h+$//g;
+        $seen_lines{$line}++;
+        next if ($seen_lines{$line} > 1); 
+        # now dedupe within the line
+        my @before_words = split(/,/, $line);
+        my %seen_words;
+        my @after_words;
+	foreach my $w (@before_words){
+	    $w =~s/^\h+//g;
+	    $w =~s/\h+$//g;
+	    $seen_words{$w}++;
+	    next if ($seen_words{$w} > 1);
+	    push(@after_words,$w);
+	}
+	$line = join(', ', @after_words);
+        push(@after_pieces,$line);
+    }
+    $out = join("\n", @after_pieces);
 
-    $out =~ s/^\s+//;
-    $out =~ s/\s+$//;
-    return $out;
+    $out =~ s/^\s+//; # remove leading whitespace
+    $out =~ s/\s+$//; # remove end whitespace
+
+    $out .= "\n";     # add final newline
+    return $out;      # we are done
 }
 
 sub _render_template {
@@ -432,6 +461,7 @@ sub _render_template {
     };
     
     my $output = $THTemplate->render($context);
+    #warn "in _render pre _clean $output";
     $output = $self->_clean($output);
 
     # is it empty?
@@ -473,7 +503,7 @@ Geo::Address::Formatter - take structured address data and format it according t
 
 =head1 VERSION
 
-version 1.51
+version 1.52
 
 =head1 SYNOPSIS
 
